from dios.dios import *
from abc import abstractmethod


class _Indexer:
    def __init__(self, _dios):
        self._dios = _dios
        self._data = _dios._data


class _LocIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        if isinstance(key, tuple):
            if len(key) > 2:
                raise KeyError("To many indexers")
            rowkey, colkey = key
        else:
            rowkey, colkey = key, slice(None)

        if isinstance(rowkey, tuple):
            raise KeyError(f"{key}. tuples are not allowed.")

        data = self._dios[colkey]

        # .loc[:, any] - simple low-cost optimization
        if isinstance(rowkey, slice) and rowkey == slice(None):
            new = data.copy()

        # .loc[any, scalar]
        elif isinstance(data, pd.Series):
            new = data.loc[rowkey]

        elif isinstance(data, self._dios.__class__):
            # .loc[scalar, any]
            if is_hashable(rowkey):
                new = pd.Series(index=type(data.columns)([]))
                for k in data.columns:
                    s = data._data.at[k]
                    new.at[k] = s.loc[rowkey]

            # .loc[non-scalar, non-scalar]
            else:
                new = self._dios.copy_empty()
                for k in data.columns:
                    s = data._data.at[k]
                    new._data.at[k] = s.loc[rowkey]
        else:
            raise AssertionError(f"getitem returned data of type {type(data)}")

        return new

    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            if len(key) > 2:
                raise KeyError("To many indexers")

            rowkey, colkey = key

            if isinstance(rowkey, tuple):
                raise KeyError(f"{key}. tuples are not allowed.")

            # .loc[-,new-scalar] = val
            # if a unknown colkey was given, we insert it and ignore rowkey
            if is_hashable(colkey) and colkey not in self._dios.columns:
                self._dios._insert(colkey, value)
                return
        else:
            rowkey, colkey = key, slice(None)

        # get .loc[any,any] - we use key(!) here
        data = self.__getitem__(key)

        if is_dios_like(value) or is_nested_list_like(value):
            raise TypeError(".loc[] cannot be used to set multi-dimensional values, use .aloc[] instead.")

        # .loc[scalar, any] or .loc[any, scalar]
        if isinstance(data, pd.Series):
            # .loc[scalar, non-scalar] - column-labeled series
            if is_hashable(rowkey):
                data.loc[:] = value
                for k in data.index:
                    s = data._data.at[k]
                    s.at[rowkey] = data.at[k]
                    self._data.at[k] = s

            # .loc[non-scalar, scalar] - (normal) row-labeled series
            elif is_hashable(colkey):
                data.loc[rowkey] = value
                self._data.at[colkey] = data

            else:
                raise AssertionError(f"getitem returned data of type {type(data)}")

        # .loc[non-scalar, non-scalar]
        elif isinstance(data, self._dios.__class__):
            for k in data.columns:
                s = data._data.at[k]
                s.loc[rowkey] = value
                self._data.at[k] = s

        # .loc[scalar, scalar]
        else:
            s = self._data.at[colkey]
            s.at[rowkey] = value
            self._data.at[colkey] = s


class _iLocIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _set_item(self, key, ix, right, ixalign=False):
        # we do use loc instead of iloc as we get real keys.
        # this works, because keys keep sorted if they come
        # from an series-index (doesn't work with df-index)
        ser = self._data.loc[key]
        if ixalign:
            ix = ser.index.intersection(ix.index)
        if isinstance(right, pd.Series):
            left = ser[ix]
            right, ix = align_index_by_policy(left, right)
        ser.iloc[ix] = right
        self._data.loc[key] = ser

    def _get_item(self, key, ix, ixalign=False):
        ser = self._data.loc[key]
        if ixalign:
            ix = ser.index.intersection(ix.index)
        return ser.iloc[ix]

    def _unpack_key(self, key):
        lowdim = False
        if isinstance(key, tuple):
            key, ckey, *fail = key
            if fail:
                raise KeyError("To many indexers")
            if is_dios_like(ckey):
                raise ValueError("Cannot index with multi-dimensional key")
            if is_bool_series(ckey):
                keys = ckey.where(ckey).dropna().index.to_list()
            else:
                if is_integer(ckey):
                    ckey, lowdim = [ckey], True
                keys = self._data.iloc[ckey].index.to_list()
        else:
            keys = self._columns.to_list()
        return keys, key, lowdim


class _aLocIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        # fallback to loc
        if not is_dios_like(value) and not is_nested_list_like(value):
            self._dios.loc[key] = value

    def __getitem__(self, item):
        return item


class _AtIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _check_key(self, key):
        if not (isinstance(key, tuple) and len(key) == 2
                and is_hashable(key[0]) and is_hashable(key[1])):
            raise KeyError(f"{key}. `.at` takes exactly one scalar row-key "
                           "and one scalar column-key")

    def __getitem__(self, key):
        self._check_key(key)
        return self._data.at[key[1]].at[key[0]]

    def __setitem__(self, key, value):
        self._check_key(key)
        if is_dios_like(value) or is_nested_list_like(value):
            raise TypeError(".at[] cannot be used to set multi-dimensional values, use .aloc[] instead.")
        s = self._data.at[key[1]]
        s.at[key[0]] = value
        self._data.at[key[1]] = s


class _iAtIndexer(_Indexer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _check_key(self, key):
        if not (isinstance(key, tuple) and len(key) == 2
                and is_integer(key[0]) and is_integer(key[1])):
            raise KeyError(f"{key} `.iat` takes exactly one integer positional "
                           f"row-key and one integer positional scalar column-key")

    def __getitem__(self, key):
        self._check_key(key)
        return self._data.iat[key[1]].iat[key[0]]

    def __setitem__(self, key, value):
        self._check_key(key)
        if is_dios_like(value) or is_nested_list_like(value):
            raise TypeError(".iat[] cannot be used to set multi-dimensional values, use .aloc[] instead.")
        s = self._data.iat[key[1]]
        s.iat[key[0]] = value
        self._data.iat[key[1]] = s


def _unpack_value(keys, ix, val):
    """Return a generator that yield (column key, corresponding value, value-align(bool) )
    for all columns.
    This is analogous to DictOfSeries._unpack_value, but with some modifications."""

    # prepare value
    val = list(val) if is_iterator(val) else val
    val = val.squeeze(axis=1) if is_dios_like(val) else val
    dioslike, nlistlike = is_dios_like(val), is_nested_list_like(val)

    # check value
    if nlistlike and len(val) != len(keys):
        raise ValueError(f"could not broadcast input array with length {len(val)}"
                         f" into dios of length {len(keys)}")
    if dioslike:
        keys = val.columns.intersection(keys)

    for i, k in enumerate(keys):
        if dioslike:
            yield k, ix, val[k]
        elif nlistlike:
            yield k, ix, val[i]
        else:
            yield k, ix, val


def maby_set_series_name(maybe_ser, name):
    if isinstance(maybe_ser, pd.Series):
        maybe_ser.name = name
